import numpy as np
from numpy import pi
from quakeio.core import QuakeComponent, QuakeSeries

try:
    import jax.numpy as jnp
except:
    import numpy as jnp


from pathlib import Path
import matplotlib.pyplot as plt
try:
    style_file = Path(__file__).parents[0]/"brace2.mplstyle"
    plt.style.use([style_file])
    import tempfile
    with tempfile.NamedTemporaryFile() as tf:
        plt.plot([0.0, 1.0], [0.0, 0.1], label=r"$\sigma")
        plt.savefig(tf.name)
except RuntimeError as e:
    import matplotlib
    matplotlib.rcParams.update({
        "text.usetex": False,
        "text.latex.preamble": ""
    })
finally:
    plt.close()



def plot_grid(series, label=None):
    fig, ax = plt.subplots(len(series),1,sharex=True)
    cycler = ax[0]._get_lines.prop_cycler
    lbl = (lambda x: x[label]) if label is not None else lambda x: None
    if label is not None and not isinstance(label,(tuple,list)):
        label = [s[label] for s in series]
    for i,s in enumerate(series):
        s.plot(ax=ax[i],label=label[i],color=next(cycler)['color'])

    if label is not None:
        fig.legend()


def _plot_func(plot):

    def __call__(self, *args, **kwds):
        if "ax" in kwds:
            self.ax = kwds.pop("ax")
            self.fig = self.ax.figure
        else:
            self.fig, self.ax = plt.subplots()

        if "title" in kwds:
            self.ax.set_title(kwds.pop("title"))

        plot(self, *args, **kwds)
        return self.ax
    return __call__


class Spectrum:
    def __init__(self, accel, *args, **kwds):
        self._accel = accel
        self.kwds = kwds

    def accel(self):
        pass

    def time(self):
        pass

    def spect(self,accel=None,dt=None,damping=None,per=None,gamma=1/2,beta=1/4,interp=None):
        if interp is None:
            from scipy.interpolate import interp1d as interp
        if damping is None:
            damping = ("damping" in self.kwds and self.kwds.pop("damping")) or 0.0

        if accel is None:
            accel = self._accel
        if isinstance(accel, QuakeComponent):
            dt = accel.accel["time_step"]
            accel = accel.accel.data
        elif isinstance(accel, QuakeSeries):
            dt = accel["time_step"]
            accel = accel.data

        if isinstance(damping, float):
            damping = [damping]

        if per is None:
            per = self.kwds.get("per",None)
        if per is None:
            per = np.arange(0.02, 1.0, 0.01)
        elif isinstance(per,tuple):
            per = np.linspace(*per)

        return _accel_spectrum(
            accel=accel,
            dt=dt,
            per=per,
            gamma=gamma,
            beta=beta,
            damping=damping,
            interp=interp
        )


    @_plot_func
    def plot(self, **kwds):
        #sa = spectrum(self._accel, **kwds)
        tsa = self.spect(self._accel, **kwds)
        if len(tsa) > 2:
            for sa in tsa[1:]:
                self.ax.plot(tsa[0], sa)
        else:
            self.ax.plot(tsa[0],tsa[1])
        self.ax.set_xlabel(f"Period, (sec.)")
        chn = self._accel["channel"]
        self.ax.set_title(f"Response spectrum (Chn. {chn})")


class TransferFunction(Spectrum):
    def __init__(self, pairs, **kwds):
        self._pairs = pairs
        self.kwds = kwds

    def tf(self):
        self.tf = tsa = transfer_function(self._pairs, **self.kwds)
        return tsa


    @_plot_func
    def plot(self, **kwds):
        self.tf = tsa = transfer_function(self._pairs, **self.kwds)
        if len(tsa) >= 2:
            for sa in tsa[1:]:
                self.ax.plot(tsa[0], sa)
        self.ax.set_xlabel(f"Period, (sec.)")
        chn1 = self._pairs[0]["channel"]
        chn2 = self._pairs[1]["channel"]
        self.ax.set_title(f"Transfer function (Chn. {chn1} vs. {chn2})")


def transfer_function(pairs, *args, **kwds):
    s1,s2 = Spectrum(pairs[0], **kwds).spect(), Spectrum(pairs[1], **kwds).spect()
    t = s2/s1
    t[0,:] = s1[0,:]
    return t

def spectrum(*args, **kwds):
    return _accel_spectrum(*args, **kwds)

def _accel_spectrum(accel,dt,damping,per,gamma=1/2,beta=1/4,interp=None):
    numper = len(per)
    m = 1.0
    numdata = len(accel)
    t = np.arange(0,(numdata)*dt,dt)

    u0,v0 = 0.0, 0.0
    SA = np.zeros((1+len(damping), numper))
    SA[0,:] = per[:]

    for di,dmp in enumerate(damping):

        for i in range(numper):
            if dt/per[i]>0.02:
                dtp = per[i]*0.02
                dtpx = np.arange(0,max(t),dtp)
                dtpx = dtpx
                accfrni = interp(t, accel)(dtpx)
                accfrn = accfrni[1:len(accfrni)-1]
                numdatan = len(accfrn)
                p  =  -m*accfrn
            else:
                dtp = dt
                accfrn = accel
                p = -m*accfrn
                numdatan = numdata

            u = np.zeros((numdatan))
            v = np.zeros((numdatan))
            a = np.zeros((numdatan))

            k = 4*pi**2*m/per[i]**2
            c = 2*dmp*np.sqrt(k/m)
            kstar = k + gamma*c/(beta*dtp) + m/(beta*dtp**2.0)
            acons = m/(beta*dtp) + gamma*c/beta
            bcons = m/(2*beta) + dtp*(gamma/(2*beta)-1)*c
            u[0] = u0
            v[0] = v0
            a[0] = (p[0]-c*v[0]-k*u[0])/m
            for j in range(1,numdatan):
                deltap = p[j]-p[j-1]
                deltaph = deltap+acons*v[j-1]+bcons*a[j-1]
                deltau = deltaph/kstar
                deltav = gamma*deltau/(beta*dtp) - gamma*v[j-1]/beta+dtp*(1-gamma/(2*beta))*a[j-1]
                deltaa = deltau/(beta*dtp**2)-v[j-1]/(beta*dtp)-a[j-1]/(2*beta)
                u[j] = u[j-1] + deltau
                v[j] = v[j-1] + deltav
                a[j] = a[j-1] + deltaa
            atot = a + accfrn
            SA[1+di,i] = abs(max(atot, key=abs))
    return SA





import folium

class AssetMap:
    file_ext = "*.json"
    def __init__(self):
        self.map = folium.Map(
            location=[37.6485, -118.9721],
            # tiles="Stamen Toner",
            tiles="Stamen Terrain",
            zoom_start=6
        )

    def vega_img(self):
        import base64
        from folium import IFrame
        encoded = base64.b64encode(open("img/a.png", 'rb').read()).decode()        
        html = '<img src="data:image/png;base64,{}">'.format
        resolution, width, height = 1, 450, 250
        iframe = IFrame(html(encoded), width=(width*resolution)+20, height=(height*resolution)+20)
        popup = folium.Popup(iframe, max_width=2650)
        
        icon = folium.Icon(color="red", icon="ok")
        return popup, icon

    def add_collection(self, collection):
        pass

    def proc_event(self, *args, **kwds):
        import quakeio
        import quakeio.vega
        record = quakeio.read("events/58658_007_20210426_10.09.54.P.zip")
        plot = folium.Vega(quakeio.vega.record2vega(record["abutment_1"]["tran"]),
                width=450, height=250
        )

        folium.Marker(
            location=[40.5031, -124.1009],
            #popup="Painter Street Overpass",
            popup=self.vega_test(),
            radius=1000,
            color="crimson",
            fill=False,
        ).add_to(self.map)
        
        folium.Marker(
            location=[37.6907, -122.0993],
            popup="Hayward Hwy 580/238 Interchange",
            radius=1000,
            color="crimson",
            fill=False,
        ).add_to(self.map)

        pop, ico = self.vega_img()
        folium.Marker(
            location=[32.7735, -115.4481],
            #popup="Meloland Overpass",
            color="crimson",
            radius=1000,
            fill=False,
            popup=pop, icon=ico
            #popup=folium.Popup(max_width=450).add_child(plot)
            #color="#3186cc",
            #fill=True,
            #fill_color="#3186cc",
        ).add_to(self.map)
 
        #return {"content": m.get_root().render()}
        #return {"content": m._repr_html_()}
        return {"content": self.map._repr_html_()}





import json
from copy import copy
from pathlib import Path

from quakeio.core import QuakeComponent, QuakeCollection
from quakeio.utils.parseutils import open_quake


class Unused:
    pass

class QuakeEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, "tolist"):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def read_json(read_file, **kwds: Unused):

    return json.load(read_file)


def record2vega(ground_motion, dump_json=True, **kwds):
    output = {
      "$schema": "https://vega.github.io/schema/vega/v5.json",
      "description": "A basic line chart example.",
      "width": 450,
      "height": 250,
      "padding": 5,

      "signals": [
        {
          "name": "interpolate",
          "value": "linear",
          "bind": {
            "input": "select",
            "options": [
              "basis",
              "cardinal",
              "catmull-rom",
              "linear",
              "monotone",
              "natural",
              "step",
              "step-after",
              "step-before"
            ]
          }
        }
      ],

      "scales": [
        {
          "name": "x",
          #"type": "point",
          "range": "width",
          "domain": {"data": "table", "field": "x"}
        },
        {
          "name": "y",
          "type": "linear",
          "range": "height",
          "nice": True,
          "zero": True,
          "domain": {"data": "table", "field": "y"}
        },
        {
          "name": "color",
          "type": "ordinal",
          "range": "category20",
          #"range": "linear",
          "domain": {"data": "table", "field": "c"}
        }
      ],

      "axes": [{"orient": "bottom", "scale": "x"}, {"orient": "left", "scale": "y"}],

      "marks": [
        {
          "type": "group",
          "from": {
            "facet": {"name": "series", "data": "table", "groupby": "c"}
          },
          "marks": [
            {
              "type": "line",
              "from": {"data": "series"},
              "encode": {
                "enter": {
                  "x": {"scale": "x", "field": "x"},
                  "y": {"scale": "y", "field": "y"},
                  "stroke": {"scale": "color", "field": "c"},
                  "strokeWidth": {"value": 2}
                },
                "update": {
                  "interpolate": {"signal": "interpolate"},
                  "strokeOpacity": {"value": 1}
                },
                "hover": {"strokeOpacity": {"value": 0.5}}
              }
            }
          ]
        }
      ]
    }
    #ground_motion = ground_motion.serialize(serialize_data=not summarize)
    dt = ground_motion.accel.time_step
    output.update({
      "data": [
        {
          "name": "table",
          "values": [
              {"x": i*dt, "y": y, "c": 0} for i, y in enumerate(ground_motion.accel)
          ]
        }
      ]
    })
    with open("aaa.json","w+") as f:
        json.dump(output,f,indent=2)
    return output

def write_json(write_file, ground_motion, indent=4, summarize=False, **kwds):
    pass

def read_yaml(read_file, **k):
    import yaml

    return yaml.load(read_file, Loader=yaml.Loader)


def write_yaml(write_file, ground_motion, summarize=False, **kwds):
    import yaml

    ground_motion = ground_motion.serialize(serialize_data=not summarize)
    # if not summarize:
    #     ground_motion = ground_motion.serialize()

    with open_quake(write_file, "w") as f:
        yaml.dump(ground_motion, f)


FILE_TYPES = {
    "vega.json": {"read": read_json, "write": write_json},
    "vega.yaml": {"read": read_yaml, "write": write_yaml}
}



# Claudio Perez
# 2021

import numpy as np
import scipy.integrate


def get_time_step(series, time_series=None, time_step: float = None) -> float:
    """
    If both `time_series` and `time_step` are None,
    `series` is checked for a `time_series` or `time_step` attribute.
    """
    if time_step:
        return time_step
    elif time_series:
        return np.diff(time_series) / len(time_series)
    else:
        if hasattr(series, "time_step"):
            return series.time_step
        elif hasattr(series, "time_series"):
            time = series.time_series
            return np.diff(time) / len(time)
        else:
            msg = "Unable to deduce time step"
            raise TypeError(msg)


def get_time_series(series, time_series, time_step):
    """
    If both `time_series` and `time_step` are None,
    `series` is checked for a `time_series` or `time_step` attribute.
    """
    if time_series:
        return time_series
    elif time_step:
        return np.cumsum([time_step] * len(series))
    else:
        if hasattr(series, "time_series"):
            return series.time_series
        elif hasattr(series, "time_step"):
            return np.cumsum([series.time_step] * len(series))
        else:
            msg = "Unable to deduce time series"
            raise TypeError(msg)


def rotate_series_in_place(series, sn=None, cs=None):
    series *= np.sin(sn or (np.pi / 2 * cs))


def arias_intensity(
    series, start_level=0.0, end_level=1.0, time_series=None, time_step=None
):
    time = get_time_series(series, time_series, time_step)
    arias_factor = np.pi / 2.0
    husid = integrate_husid(series, time)
    # Normalize
    husid_norm = husid / husid[-1]
    idx = np.where(np.logical_and(husid_norm >= start_level, husid_norm <= end_level))[
        0
    ]
    if len(idx) < len(series):
        husid = integrate_husid(series[idx], time)

    raise NotImplementedError("Not implemented")
    return arias_factor * husid[-1]


def integrate_husid(series, time_series=None, time_step=None):
    time = get_time_series(series, time_series, time_step)
    return scipy.integrate.cumtrapz(series ** 2.0, time)




import numpy as np

from quakeio.core import QuakeSeries, QuakeComponent
from quakeio.utils.parseutils import open_quake


def read(read_file, **kwds):
    with open_quake(read_file, "r") as f:
        time_step = next(next(f))
        accel = np.genfromtxt(f, skip_header=1, delimiter=",", usecols=0)
    return QuakeComponent(
        QuakeSeries(accel), meta={"accel.time_step": time_step}
    )


def write(write_file, motion, label="eqsig data", **kwds):
    if isinstance(motion,QuakeSeries):
        accel = motion
    elif isinstance(motion,QuakeComponent):
        accel = motion.accel
    time_step = accel["time_step"]
    data = [label, f"{len(accel)} {time_step}", *map(str, accel.data)]
    with open_quake(write_file, "w") as f:
        f.write("\n".join(data))


FILE_TYPES = {"eqsig": {"read": read, "write": write}}




# Claudio Perez
# 2021
import json
from copy import copy
from pathlib import Path
import warnings
from enum import Enum

#import opensees.units
import numpy as np

class st(Enum): ACCEL, DISPL, VELOC = range(3)
class dt(Enum): LONG,  TRAN,  VERT  = range(3)
class pt(Enum): SECT,  PLAN,  ELEV  = range(3)


DIRECTIONS = ["long", "tran", "up"]


class QuakeCollection(dict):
    """
    Container for a collection of QuakeMotion objects
    """

    def __init__(self, motions, event_date=None, meta=None, **kwds):
        meta = meta if meta is not None else {}
        dict.__init__(self, **meta)

        self.motions = motions
        self.event_date = event_date
        for motion in motions.values():
            motion._parent = self
            for comp in motion.components.values():
                comp._parent = motion

    def __repr__(self):
        return f"QuakeCollection({dict.__repr__(self)})"

    @property
    def components(self):
        for motion in self.motions.values():
            for component in motion.components.values():
                yield component

    def serialize(self, serialize_data=True, **kwds) -> dict:
        data = {k: v for k, v in self.items()}
        return {**data, "motions": [i.serialize(**kwds) for i in self.motions.values()]}

    @property
    def components(self):
        for m in self.motions.values():
            for v in m.components.values():
                yield v

    def convert_units(self, units):
        for component in self.components:
            component.convert_units(units)

    def match(self, t, *args, **kwds):
        if callable(t):
            pass

        assert type(t) == str

        if t[0] == "l":
            return self.at(*args, **kwds)
        if t[0] == "r":
            import re
            test = lambda x, y: re.match(x, y)
        elif t == ">":
            test = lambda x, y: x > y
        elif t == "<":
            test = lambda x, y: x < y

        if len(t) > 1:
            typ = "multi"
        else:
            typ = "single"

        if typ=="single":
            for motion in self.motions.values():
                tests = (
                  k in motion and test(v, motion[k]) for k, v in kwds.items()
                )
                if all(tests):
                    return motion
                for component in motion.components.values():
                    if all(k in component and test(v,component[k]) for k, v in kwds.items()):
                        return component
        elif typ=="multi":
            res = []
            for motion in self.motions.values():
                tests = (
                  k in motion and test(v, motion[k]) for k, v in kwds.items()
                )
                if all(tests):
                    res.append(motion)
                for component in motion.components.values():
                    if all(k in component and test(v,component[k]) for k, v in kwds.items()):
                        res.append(component)
            return res

        return self.at(*args, **kwds)

    def at(self, rtol=1e-05, atol=1e-08, **kwds):
        for motion in self.motions.values():
            tests = (
              k in motion and (
                np.isclose(motion[k],v, rtol=rtol, atol=atol) if isinstance(v,float)\
                        else (motion[k] == v)
              ) for k, v in kwds.items()
            )
            if all(tests):
                return motion

        return self.get_component(**kwds)

    def get_component(self, **kwds):
        for motion in self.motions.values():
            for component in motion.components.values():
                if all(component[k] == v for k, v in kwds.items()):
                    return component


class QuakeMotion(dict):
    """
    A container of `QuakeComponent` objects.
    """
    def _update_components(f):
        def wrapped(*args, **kwds):
            res = f(*args, **kwds)
            [getattr(cmp,s)._refresh() 
                    for cmp in res.components.values()
                        for s in ["accel","veloc","displ"] if hasattr(cmp,s)]
            return res
        return wrapped

    def __init__(self, components: dict = None, meta:dict=None):
        dict.__init__(self)
        self.directions = ["long", "tran", "up"]
        self.components = components if components is not None else {}
        #for k,v in self.components.items():
        #    setattr(self,k,v)
        self.update(meta if meta is not None else {})
        for comp in self.components.values():
            comp._parent = self

    @property
    def long(self):
        return self.components["long"]
    @property
    def tran(self):
        return self.components["tran"]

    @property
    def accel(self):
        return np.stack(tuple(c.accel for c in self.components.values())).T

    @property
    def veloc(self):
        return np.stack(tuple(c.veloc for c in self.components.values())).T

    @property
    def displ(self):
        return np.stack(tuple(c.displ for c in self.components.values())).T

    def __repr__(self):
        #return f"QuakeMotion({dict.__repr__(self)})"
        return f"QuakeMotion({dict.__repr__(self)})"
    
    def __sub__(self, other):
        ret = copy(self)
        ret.components = {}
        for dirn in DIRECTIONS:
            if dirn in other.components and dirn in self.components:
                ret.components[dirn] = self.components[dirn] - other.components[dirn] 
        return ret
    
    def __add__(self, other):
        ret = copy(self)
        ret.components = {}
        for dirn in DIRECTIONS:
            if dirn in other.components and dirn in self.components:
                ret.components[dirn] = self.components[dirn] + other.components[dirn] 

        return ret
    
    def __rsub__(self, other):
        return self.__sub__(other)
    
    def __radd__(self, other):
        return self.__add__(other)


    def serialize(self, serialize_data=True, **kwds) -> dict:
        return {
            **self,
            "components": [c.serialize(**kwds) for c in self.components.values()]
        }

    @_update_components
    def slice(self, *args):
        for c in self.components.values():
            c.slice(*args)
        return self

    @_update_components
    def rotate(self, angle=None, rotation=None, vert=None):
        """
        > NOTE: This method changes data in the class instance.
        """

        if vert == 3:
            angle *= -1

        rx, ry = (
            np.array([[ np.cos(angle),-np.sin(angle)], 
                      [ np.sin(angle), np.cos(angle)]])
            if not rotation else rotation
        )

        try:
            for attr in ["accel", "veloc", "displ"]:
                x = getattr(self.components["long"], attr).data
                y = getattr(self.components["tran"], attr).data
                X = np.array([x, y])
                x[:] = np.dot(rx, X)
                y[:] = np.dot(ry, X)

        except KeyError as e:
            raise AttributeError("Attempt to rotate a motion that"\
                    f"does not have a '{e.args[0]}' component")

        return self

    def resultant(self):
        # initialize
        series = {k:None for k in ("accel", "veloc", "displ")}
        for typ in series:
            try:
                series[typ] = np.sqrt(sum(
                    np.power(getattr(self.components[dirn],typ), 2)
                    for dirn in self.directions 
                        if dirn in self.components and self.components[dirn]
                ))
            except AttributeError as e:
                raise e
        return QuakeComponent(*series.values())

    def match(self, t, *args, **kwds):
        if callable(t):
            pass

        assert type(t) == str

        if t[0] == "l":
            return self.at(*args, **kwds)
        if t[0] == "r":
            import re
            test = lambda x, y: re.match(x, y)
        elif t == ">":
            test = lambda x, y: x > y
        elif t == "<":
            test = lambda x, y: x < y

        if len(t) > 1:
            typ = "multi"
        else:
            typ = "single"

        if typ=="single":
            for motion in self.components.values():
                tests = (
                  k in motion and test(v, motion[k]) for k, v in kwds.items()
                )
                if all(tests):
                    return motion

        elif typ=="multi":
            res = []
            for motion in self.components.values():
                tests = (
                  k in motion and test(v, motion[k]) for k, v in kwds.items()
                )
                if all(tests):
                    res.append(motion)
            return res

        return self.at(*args, **kwds)

    def at(self, rtol=1e-05, atol=1e-08, **kwds):
        print(kwds)
        for motion in self.components.values():
            tests = (
              k in motion and (
                np.isclose(motion[k],v, rtol=rtol, atol=atol) if isinstance(v,float)\
                        else (motion[k] == v)
              ) for k, v in kwds.items()
            )
            if all(tests):
                return motion

        return []


class QuakeComponent(dict):
    """A container of QuakeSeries objects"""
    schema_dir = Path(__file__).parents[2] / "etc/schemas"
    schema_file = schema_dir / "component.schema.json"

    def __init__(self, accel, veloc, displ, motion=None, meta=None):
        meta = meta if meta is not None else {}
        self.accel = accel
        self.displ = displ
        self.veloc = veloc
        self._parent = motion

        for series in [accel, veloc, displ]:
            series._parent = self

        dict.__init__(self, **meta)

    @property
    def series(self):
        for s in "accel","veloc","displ":
            yield getattr(self,s)

    def __repr__(self):
        return f"QuakeComponent({self['file_name']}) at {hex(id(self))}"

    def find_components(self):
        loc = self["location_name"]
        for comp in self._parent.components.values():
            if comp["location_name"] == loc:
                yield comp

    def slice(self, *args):
        for s in "accel","veloc","displ":
            getattr(self,s).slice(*args)
        return self

    def serialize(
        self,
        ljust=0,
        serialize_data=True,
        serialize_series=True,
        humanize_keys=False,
        **kwds,
    ) -> dict:
        if humanize_keys:
            key_name = get_schema_key_map(self.schema_file)
            ret = {key_name(k).ljust(ljust, "."): val for k, val in self.items()}
        else:
            ret = dict(self)
        if serialize_series:
            ret.update(
                {
                    "accel": self.accel.serialize(
                        "accel", serialize_data=serialize_data, **kwds
                    ),
                    "veloc": self.veloc.serialize(
                        "veloc", serialize_data=serialize_data, **kwds
                    ),
                    "displ": self.displ.serialize(
                        "displ", serialize_data=serialize_data, **kwds
                    ),
                }
            )
        return ret



    def __sub__(self, other):
        ret = copy(self)
        if isinstance(other, QuakeComponent):
            for k in ["accel", "veloc", "displ"]:
                setattr(ret, k, getattr(other, k) - getattr(self, k))

        elif isinstance(other, (float, int)):
            for k in ["accel", "veloc", "displ"]:
                setattr(ret, k, other - getattr(self, k))
        return ret

    def __add__(self, other):
        ret = copy(self)
        if isinstance(other, QuakeComponent):
            for k in ["accel", "veloc", "displ"]:
                setattr(ret, k, getattr(other, k) + getattr(self, k))

        elif isinstance(other, (float, int)):
            for k in ["accel", "veloc", "displ"]:
                setattr(ret, k, other + getattr(self, k))
        return ret

    def __mul__(self, other):
        ret = copy(self)
        if isinstance(other, QuakeComponent):
            for k in ["accel", "veloc", "displ"]:
                setattr(ret, k, getattr(other, k) * getattr(self, k))
        elif isinstance(other, (float, int)):
            for k in ["accel", "veloc", "displ"]:
                setattr(ret, k,  other * getattr(self, k))
        return ret

    def __rsub__(self, other):
        return self.__sub__(other)

    def __radd__(self, other):
        return self.__add__(other)

    def __rmul__(self, other):
        return self.__mul__(other)


class QuakeSeries(dict):
    def _update_metadata(f):
        def wrapped(*args, **kwds):
            res = f(*args, **kwds)
            res._refresh()
            return res
        return wrapped

    def __init__(self, input_array, dt=None, meta=None, time_zero=0.0, **kwds):
        self._data = np.asarray(input_array)
        assert len(self.data.shape) == 1
        self.update(meta if meta is not None else {})
        self.update(kwds)
        self._time = None
        self.time_zero = time_zero
        if dt is not None:
            self.time_step = self["time_step"] = dt
        if "peak_value" not in self:
            self._refresh()

    def __getitem__(self, key):
        try:
            return dict.__getitem__(self,key)
        except KeyError as e:
            if hasattr(self, "_parent"):
                return self._parent[key]
            else:
                raise e

    def _refresh(self):
        if len(self.data) > 0:
            self["peak_value"] = max(self.data, key=abs)
        return self

    @property
    def data(self):
        return self._data

    def __repr__(self):
        try:
            filename=hex(id(self))
            return f"QuakeSeries({filename},{self['units']})" #{dict.__repr__(self)})"
        except:
            return f"QuakeSeries([{self.data[0]}...{self.data[-1]}])"
    
    def serialize(
        self, key=None, summarize=False, serialize_data=True, humanize_keys=False
    ):

        attributes = {
            #".".join((key, k)): v for k,v in self.items() if k[0] != "_"
            k: v for k,v in self.items() if k[0] != "_"
        }
        if serialize_data and not summarize:
            attributes.update({"data": list(self.data)})
        return attributes

    @_update_metadata
    def __pow__(self, other):
        ret = copy(self)
        ret._data = np.power(ret.data, other)
        return ret

    @_update_metadata
    def __sub__(self, other):
        ret = copy(self)
        if isinstance(other, QuakeSeries):
            ret._data = other.data - self.data

        elif isinstance(other, (float, int)):
            ret._data = other - self.data
        return ret

    @_update_metadata
    def __add__(self, other):
        ret = copy(self)
        if isinstance(other, QuakeSeries):
            ret._data += other.data

        elif isinstance(other, (float, int)):
            ret._data += other
        return ret

    @_update_metadata
    def __mul__(self, other):
        if isinstance(other, (float, int)):
            ret = copy(self)
            ret._data *= other
        else:
            raise TypeError()
        return ret

    @_update_metadata
    def sqrt(self, inplace=False):
        ret = copy(self)
        out = self._data if inplace else None
        ret._data = np.sqrt(self._data, out=out)
        return ret

    def __array__(self,dtype=None):
        return self._data
    
    def __array_ufunc__(self, ufunc, method, *inputs, **kwds):
        if method=="__call__":
            inputs = [i.data if isinstance(i,self.__class__) else i for i in inputs]
            return self.__class__(ufunc(*inputs, **kwds),meta=self,time_zero=self.time_zero)._refresh()
    
    def __rsub__(self, other):
        return self.__sub__(other)
    
    def __radd__(self, other):
        return self.__add__(other)

    def __rmul__(self, other):
        return self.__mul__(other)

    @property
    def time(self):
        if self._time is None:
            dt = self["time_step"]
            t0 = self.time_zero
            self._time = np.arange(t0,t0+dt*len(self.data),dt)
        return self._time

    def slice(self,*args):
        time = self.time
        idx = (args[0] < time) & (time < args[1])
        self._time = time[idx]
        self.time_zero = self._time[0]
        self._data = self.data[idx]
        return self

    def plot(self, ax=None, fig=None, label=None, index=(None,), scale=1.0, **kwds):
        # matplotlib takes a very long time to import; avoid
        # loading at module level.
        import matplotlib.pyplot as plt
        idx = slice(*index)
        time = self.time

        if ax is None:
            fig, ax = plt.subplots()
        if label in self:
            label=self[label]
        if label is not None:
            kwds["label"] = label
        ax.plot(time[idx],self.data[idx]*scale, **kwds)
        ax.set_ylabel(f"({self['units']})")
        ax.set_xlabel("time (sec.)")
        return ax



def rotate(data, angle):
    #output = copy(data)
    if isinstance(data, QuakeComponent):
        try:
            data._parent.rotate(angle)
        except Exception as e:
            raise Exception(f"Unable to rotate single component ({e})")

    elif isinstance(data, QuakeCollection):
        for name, record in data.items():
            try:
                record.rotate(angle)
            except KeyError:
                warnings.warn(f"Not rotating record {name}")


def get_schema_key_map(schema_file):
    with open(schema_file, "r") as f:
        schema = json.load(f)["properties"]

    def get_name(key):
        return schema[key]["title"] if key in schema else key

    return get_name


def write_pretty(data):
    output = copy(data)
    schema_dir = Path(__file__).parents[2] / "etc/schemas"
    schema_file = schema_dir / "component.schema.json"
    with open(schema_file, "r") as f:
        schema = json.load(f)["properties"]
    if isinstance(data, QuakeComponent):
        for k in data:
            if k in schema and "units" not in k:
                output[schema[k]["title"]] = output.pop(k)
            else:
                pass

    elif isinstance(data, QuakeCollection):
        for name, record in data.items():
            output[name] = copy(record)
            for dirn, component in record.items():
                output[name][dirn] = copy(component)
                for k, v in component.items():
                    if k in schema and "units" not in k:
                        output[name][dirn][schema[k]["title"]] = output[name][dirn].pop(
                            k
                        )
                    else:
                        pass

    return output



# Claudio Perez
__version__ = "0.1.18"

from pathlib import Path

from . import csmip, nga, eqsig, basic_formats, opensees

FILE_TYPES = {}

# Avoid repeated dot lookup
_register_file_type = FILE_TYPES.update
_register_file_type(nga.FILE_TYPES)
_register_file_type(csmip.FILE_TYPES)
_register_file_type(eqsig.FILE_TYPES)
_register_file_type(basic_formats.FILE_TYPES)
_register_file_type(opensees.FILE_TYPES)

# Map different file extensions to the name
# of the parser that should be used by default
# (when format argument is not explicitly passed).
DEFAULT_TYPES = {
    ".at2":  "nga.at2",
    ".AT2":  "nga.at2",
    ".zip":  "csmip.zip",
    ".v2":   "csmip.v2",
    ".V2":   "csmip.v2",
    ".json": "json",
}


def read(read_file, input_format=None, **kwds):
    """
    Generic ground motion reader
    """
    input_format = input_format or kwds.pop("format",None)
    if input_format is not None:
        typ = input_format
    else:
        try:
            typ = DEFAULT_TYPES[Path(read_file).suffix.lower()]
        except KeyError:
            raise ValueError("Unable to deduce input format.\n")
    return FILE_TYPES[typ]["read"](read_file, **kwds)


def write(write_file, ground_motion, write_format: str = None, *args, **kwds):
    """
    Generic ground motion writer
    """
    if write_format:
        typ = write_format
    else:
        try:
            typ = DEFAULT_TYPES[Path(write_file).suffix.lower()]
        except KeyError:
            raise ValueError("Unable to deduce output format")
    FILE_TYPES[typ]["write"](write_file, ground_motion, *args, **kwds)

def dumps(ground_motion, **kwds):
    return basic_formats.dumps_json(ground_motion, **kwds)




import re
import io
from os import PathLike
from pathlib import Path
from typing import Union, IO, Callable
import contextlib

# Regular expression for extracting decimal number
RE_DECIMAL = "[-]?[0-9]*[.]?[0-9]*"
# Regular expression for extracting units
RE_UNITS = "[A-z,/,0-9]*"

CRE_WHITE = re.compile("\\W+")


def maybe_t(pattern: str, typ: Callable, strict: bool = True):
    cpattern = re.compile(pattern, re.IGNORECASE)
    def proc(match):
        if match:
            return typ(re.search(cpattern, match).group(1))
        else:
            return None

    return proc


@contextlib.contextmanager
def open_quake(file, mode=None, archive=None):
    import sys

    if file == "-":
        if mode is None or mode == "" or "r" in mode:
            fh = sys.stdin
        else:
            fh = sys.stdout
    elif not archive and isinstance(file, (PathLike, str)):
        fh = open(file, mode)
    elif archive:
        fh = archive.open(file, mode)
    else:
        fh = file
    # if file != "-":
    #     yield fh
    #     fh.close()
    try:
        yield fh

    finally:
        if isinstance(file, str) and file != "-":
            fh.close()


def parse_sequential_fields(data, field_spec: dict, parsed_fields={}, verbose=False) -> dict:
    field_iterator = iter(field_spec.items())
    fields, (typs, pat) = next(field_iterator)
    #print(f"\tfields: {fields}")
    for ln, line in enumerate(data):
        #if not ln % 100: print(f"\t\tline: {ln}")
        if hasattr(pat, "findall"):
            match = pat.findall(str(line))
            if match:
                prefix = ""
                assert len(fields) == len(typs) == len(match[0]), (pat, match[0])
                for field, typ, val in zip(fields, typs, match[0]):
                    if field[0] == ".":
                        key = prefix + field
                    else:
                        prefix = field
                        key = field

                    parsed_fields[key] = typ(val)
                try:
                    fields, (typs, pat) = next(field_iterator)
                    #print(f"\tfields: {fields}")
                except StopIteration:
                    # All fields have been parsed out
                    break
            else:
                pass
        else:
            lnum, line_pat = pat
            lnum -= 1
            assert ln <= lnum
            if ln < lnum: continue

            prefix = ""
            line_str = str(line)
            if isinstance(line_pat, tuple):
                match = ( [line_str[s] for s in line_pat] , )
            else:
                match = line_pat.findall(line_str)
                #assert len(fields) == len(typs) == len(match[0]), (line_pat, match[0])
            if match:
                for field, typ, val in zip(fields, typs, match[0]):
                    if field[0] == ".":
                        key = prefix + field
                    else:
                        prefix = field
                        key = field
                    parsed_fields[key] = typ(val)
            fields, (typs, pat) = next(field_iterator)

    if verbose:
        print(fields)

    return parsed_fields


def get_file_type(
    file: Union[str, Path, IO], file_type: str, module: str = None
) -> str:
    if isinstance(file, io.IOBase):
        return file_type



#!/bin/env python
from pathlib import Path

import quakeio

csmip_archive = Path("dat/58658_007_20210426_10.09.54.P.zip")
csmip_dir = Path("dat/58658_007_20210426_10.09.54.P/")

#----------------------------------------------------------------------
# Record (.v2)
#----------------------------------------------------------------------
def test_read_event():
    return quakeio.csmip.read_event(csmip_archive)

def test_unique():
    event = test_read_event()
    all_components = [c for m in event.motions.values() for c in m.components]
    #all_components = [m for m in event.motions]# for c in m.components]
    assert len(all_components) == 20

#----------------------------------------------------------------------
# Record (.v2)
#----------------------------------------------------------------------
def test_2():
    csmip_record = quakeio.csmip.read_record(csmip_dir / "chan001.v2")
    quakeio.write("-", csmip_record, "json")


def test_read():
    csmip_record = quakeio.read(csmip_dir / "chan001.v2")
    quakeio.write("-", csmip_record, "yaml", summarize=True)
    return csmip_record


def test_peak():
    csmip_motion = test_read()
    assert csmip_motion.accel["peak_value"] == 17.433
    assert csmip_motion.veloc["peak_value"] == 0.205
    assert csmip_motion.displ["peak_value"] == -0.004


def test_peak_time():
    csmip_motion = test_read()
    assert csmip_motion.accel["peak_time"] == 20.270
    assert csmip_motion.veloc["peak_time"] == 20.290
    assert csmip_motion.displ["peak_time"] == 20.270

def test_accel_data():
    csmip_record = test_read()
    assert csmip_record.accel.data[0]  == -0.000102
    assert csmip_record.accel.data[-1] ==  0.000105

def test_veloc_data():
    csmip_record = test_read()
    assert csmip_record.veloc.data[0]  == 0.0000950
    assert csmip_record.veloc.data[-1] == 0.0001009


# csmip_record = quakeio.QuakeComponent(csmip_dir/"chan001.v2")
if __name__ == "__main__":
    import sys, yaml
    from pathlib import Path
    from collections import defaultdict

    file_list = sys.argv[1:]

    fields = {k: defaultdict(lambda: 0) for k in ("collection", "motion", "component", "series")}

    
    counts = defaultdict(lambda: 0)
    for file in file_list:
        try:
            collection = quakeio.read(file)
        except:
            print(file)
            continue

        counts["collection"] += 1
        for k,v in collection.items():
            fields["collection"][k] += 1
            for motion in collection.motions.values():
                counts["motion"] += 1
                for k,v in motion.items():
                    fields["motion"][k] += 1
                    for component in motion.components.values():
                        counts["component"] += 1
                        for k,v in component.items():
                            fields["component"][k] += 1
                            for series in "accel", "veloc", "displ":
                                counts["series"] += 1
                                for k,v in getattr(component, series).items():
                                    fields["series"][k] += 1

    results = {
        f"{typ} ({counts[typ]})": {field: fields[typ][field] / counts[typ] for field in fields[typ]}
        for typ in counts
    }
    yaml.dump(results, sys.stdout)
        








import numpy as np

import quakeio

def test_mul():
    from .test_csmip import test_read_event
    event = test_read_event()

def test_rotate():
    # pick an arbitrary time
    time = 23
    angle = 21/7/3
    rotation = np.array([
        [ np.cos(angle),-np.sin(angle)],
        [ np.sin(angle), np.cos(angle)]
    ])

    # read event
    from .test_csmip import test_read_event
    event = test_read_event()
    assert type(event) == quakeio.core.QuakeCollection

    mot = event.motions["bent_4_north_column_grnd_level"]
    assert type(mot) == quakeio.core.QuakeMotion


    a_long, a_tran = mot.long.accel.data[time], mot.tran.accel.data[time]
    
    # Apply rotation to record
    mot.rotate(angle) 
    A_long_rot, A_tran_rot = mot.long.accel, mot.tran.accel

    # Ensure that rotation does not alter object types
    assert type(A_long_rot) == quakeio.core.QuakeSeries
    assert type(A_tran_rot) == quakeio.core.QuakeSeries

    # manually calculate values 
    a_calc_long, a_calc_tran = rotation @ np.array([[a_long],[a_tran]])

    # Assert rotation matches manual calc
    assert np.isclose(float(A_long_rot.data[time]), float(a_calc_long))
    assert np.isclose(float(A_tran_rot.data[time]), float(a_calc_tran))


def test_rotate_displ():
    # pick an arbitrary time
    time = 23
    angle = 21/7/3
    rotation = np.array([
        [ np.cos(angle),-np.sin(angle)],
        [ np.sin(angle), np.cos(angle)]
    ])

    # read event
    from .test_csmip import test_read_event
    event = test_read_event()
    assert type(event) == quakeio.core.QuakeCollection

    mot = event.motions["bent_4_north_column_grnd_level"]
    assert type(mot) == quakeio.core.QuakeMotion


    a_long, a_tran = mot.long.displ.data[time], mot.tran.displ.data[time]
    
    # Apply rotation to record
    mot.rotate(angle) 
    D_long_rot, D_tran_rot = mot.long.displ, mot.tran.displ

    # Ensure that rotation does not alter object types
    assert type(D_long_rot) == quakeio.core.QuakeSeries
    assert type(D_tran_rot) == quakeio.core.QuakeSeries

    # manually calculate values 
    a_calc_long, a_calc_tran = rotation @ np.array([[a_long],[a_tran]])

    # Assert rotation matches manual calc
    assert np.isclose(float(D_long_rot.data[time]), float(a_calc_long))
    assert np.isclose(float(D_tran_rot.data[time]), float(a_calc_tran))




try:
    import jax.numpy as jnp
except:
    import numpy as jnp

import numpy as np
from numpy import pi
import matplotlib.pyplot as plt
from quakeio.core import QuakeComponent, QuakeSeries

def _plot_func(plot):
    plt.style.use("berkeley")

    def __call__(self, *args, **kwds):
        if "ax" in kwds:
            self.ax = kwds.pop("ax")
            self.fig = self.ax.figure
        else:
            self.fig, self.ax = plt.subplots()
            
        if "title" in kwds:
            self.ax.set_title(kwds.pop("title"))

        plot(self, *args, **kwds)
        return self.ax
    return __call__


class Spectrum:
    def __init__(self, accel, *args, **kwds):
        self._accel = accel
        self.kwds = kwds

    def accel(self):
        pass

    def time(self):
        pass

    def spect(self,accel=None,dt=None,damping=None,per=None,gamma=1/2,beta=1/4,interp=None):
        if interp is None:
            from scipy.interpolate import interp1d as interp
        if damping is None:
            damping = ("damping" in self.kwds and self.kwds.pop("damping")) or 0.0
        if isinstance(accel, QuakeComponent):
            dt = accel.accel["time_step"]
            accel = accel.accel.data
        elif isinstance(accel, QuakeSeries):
            dt = accel["time_step"]
            accel = accel.data
        if isinstance(damping, float):
            damping = [damping]

        if per is None:
            per = np.arange(0.02, 1.0, 0.01)

        return _accel_spectrum(
            accel=accel,
            dt=dt,
            per=per,
            gamma=gamma,
            beta=beta,
            damping=damping,
            interp=interp
        )
    

    @_plot_func
    def plot(self, **kwds):
        #sa = spectrum(self._accel, **kwds)
        tsa = self.spect(self._accel, **kwds)
        if len(tsa) > 2:
            for sa in tsa[1:]:
                self.ax.plot(tsa[0], sa)
        self.ax.set_title(self._accel["location_name"])


class TransferFunction:
    pass


def transfer_function(pairs, *args, **kwds):
    s1,s2 = spectrum(pairs[0], **kwds), spectrum(pairs[1], **kwds)
    t = s2/s1
    t[0,:] = s1[0,:]
    return t

def spectrum(*args, **kwds):
    return _accel_spectrum(*args, **kwds)

def _accel_spectrum(accel,dt,damping,per,gamma=1/2,beta=1/4,interp=None):
    numper=len(per)
    m = 1.0
    numdata=len(accel)
    t=np.arange(0,(numdata)*dt,dt)
    
    u0,v0 = 0.0, 0.0
    SA = np.zeros((1+len(damping),numper))
    SA[0,:] = per[:]

    for di,dmp in enumerate(damping):
        for i in range(numper):
            if dt/per[i]>0.02:
                dtp = per[i]*0.02
                dtpx = np.arange(0,max(t),dtp)
                dtpx = dtpx
                accfrni = interp(t, accel)(dtpx)
                accfrn = accfrni[1:len(accfrni)-1]
                numdatan = len(accfrn)
                p = -m*accfrn
            else:
                dtp=dt;
                accfrn = accel;
                p = -m*accfrn;
                numdatan=numdata;

            u = np.zeros((numdatan))
            v = np.zeros((numdatan))
            a = np.zeros((numdatan))

            k = 4*pi**2*m/per[i]**2;
            c = 2*dmp*np.sqrt(k/m);
            kstar = k + gamma*c/(beta*dtp) + m/(beta*dtp**2.0)
            acons = m/(beta*dtp) + gamma*c/beta
            bcons = m/(2*beta) + dtp*(gamma/(2*beta)-1)*c
            u[0] = u0
            v[0] = v0
            a[0] = (p[0]-c*v[0]-k*u[0])/m
            for j in range(1,numdatan):
                deltap = p[j]-p[j-1]
                deltaph = deltap+acons*v[j-1]+bcons*a[j-1]
                deltau = deltaph/kstar
                deltav = gamma*deltau/(beta*dtp)-gamma*v[j-1]/beta+dtp*(1-gamma/(2*beta))*a[j-1]
                deltaa = deltau/(beta*dtp**2)-v[j-1]/(beta*dtp)-a[j-1]/(2*beta)
                u[j] = u[j-1] + deltau
                v[j] = v[j-1] + deltav
                a[j] = a[j-1] + deltaa
            atot = a + accfrn
            SA[1+di,i] = abs(max(atot, key = abs));
    return SA




import quakeio
from quakeio.__main__ import main
import subprocess


def test_summarize_zip():
    main(["-S", "dat/58658_007_20210426_10.09.54.P.zip"])

def test_rotate():
    p = subprocess.Popen([
            "quakeio", 
            "-r", "2.0", 
            "-m", "station_channel=20",
            "dat/58658_007_20210426_10.09.54.P.zip"
        ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
    )

    #stdout,stderr = p.communicate()
    motion = quakeio.read(p.stdout, format="json")
    print(len(list(motion.keys())))

if __name__=="__main__":
    #test_summarize_zip()
    test_rotate()



from quakeio.core import QuakeSeries, QuakeComponent, QuakeCollection
from quakeio.utils.parseutils import open_quake
import numpy


def read(read_file, **kwds):
    with open_quake(read_file) as f:
        pass


def write(write_file, ground_motion, **kwds):
    # print(type(ground_motion))
    if isinstance(ground_motion, QuakeComponent):
        dt = ground_motion.accel["time_step"]
        accel = numpy.asarray(ground_motion.accel.data)[:, None]
    elif isinstance(ground_motion, QuakeSeries):
        dt = ground_motion["time_step"]
        accel = numpy.asarray(ground_motion.data)[:, None]
    else:
        raise ValueError(
            f"Cannot conver motion of type `{type(ground_motion)}` to QuakeComponent"
        )
    with open_quake(write_file, "w") as f:
        #numpy.savetxt(f, accel, fmt="%.8f")  # delimiter="\n")
        f.write(f"-dt {dt} -values {{")
        numpy.savetxt(f, accel, fmt="%.8f", delimiter=" ")
        f.write("}")

def write_tcl(write_file, ground_motion, **kwds):
    if isinstance(ground_motion, QuakeComponent):
        series = ground_motion.accel
    elif isinstance(ground_motion, QuakeSeries):
        series = ground_motion
    else:
        raise ValueError(
            f"Cannot conver motion of type `{type(ground_motion)}` to QuakeComponent"
        )
    with open_quake(write_file, "w") as f:
        f.write(f"dt {series['time_step']} ")
        for k,v in series.items():
            if isinstance(v,(str,int,float)):
                f.write(f"{k} {v} ")
        f.write(f"values {{")
        numpy.savetxt(f, series.data, fmt="%.8f", delimiter=" ")
        f.write("}")


FILE_TYPES = {
    "opensees": {"write": write, "type": "any"},
    "tcl":      {"write": write_tcl, "type": "any"}
}



# Claudio Perez
# Summer 2021
"""
Parse a CSMIP strong motion data file.
"""
import re
import sys
import fnmatch
import zipfile
from datetime import datetime
from pathlib import Path
from collections import defaultdict

import numpy as np

from .core import (
    QuakeCollection,
    QuakeMotion,
    QuakeComponent,
    QuakeSeries,
)

from .utils.parseutils import (
    parse_sequential_fields,
    open_quake,
    RE_DECIMAL,  # Regular expression for extracting decimal values
    RE_UNITS,    # Regular expression for extracting units
    CRE_WHITE,
    maybe_t,
)

re_digits = re.compile(r"([0-9]+)")

# Module constants
NUM_COLUMNS = 8
HEADER_END_LINE = 45

INTEGER_HEADER_START_LINE = 25
INTEGER_HEADER_END_LINE = 25+7
REAL_HEADER_END_LINE = 45

DATEFMT = "%m/%d/%y, %H:%M" #:%S.%f %Z"

# A utility to convert general strings into
# appropriate keys.
_make_key = lambda strng: strng.strip().replace(" ", "_").lower()

# Definition of some type constructors, used to coerce
# the result of a regex match.
words  = lambda x: CRE_WHITE.sub(" ", str(x)).strip()
units  = lambda x: str(x).strip().lower()

class FindDate:
    pat1  = re.compile(r"([A-z]* of )? ([A-z]{,5} [A-z]{,5} [0-9]{,2}, [0-9]{4} [0-9]{2}:[0-9]{2}:"+RE_DECIMAL+" [A-Z]{,4})")
    def findall(self, line):
        if line[0].upper() == "R":
            return FindDate.pat1.findall(line)
        else:
            print(line)
            match = FindDate.pat2.findall(line)
            print(match)
            return (["", *match[0]],)

# PARSE TABLE
# ----------------------------------------------------------------
# The parse table defines all of the fields than can be extracted
# from a CSMIP file, and maps these fields to (1) a data type and 
# (2) regular expression. This table has the following structure:
#
#   Dict[
#     Tuple[*[Key]],
#     Tuple[ Tuple[*[Type]], RegExp]
#   ]
#
# leading key (ie key.split(".")[0]) determines whether
# field is part of record or accel/veloc/displ series.
# fmt: off
HEADER_FIELDS = {
    # line 1
    ("_", "record.record_identifier"): ((str, str),
        re.compile(r"([A-z ]*) ([A-z0-9\-\.]*)", re.IGNORECASE)
    ),
    # line 5
    ("_", "record.date"): ((str,lambda s: datetime.strptime(s.strip(),DATEFMT).isoformat(),),
        #(5,
            re.compile(r"(.*): *([0-9]{,2}/[0-9]{,2}/[0-9]{,4}, *[0-9]{2}:[0-9]{2})")#[:.0-9]{,4} [A-Z]{,4})")
        #)
    ),
    # line 6
    ("record.station.no", "record.station.coord"): ((str, str),
        re.compile(
            rf"Station No\. *([0-9]*) *({RE_DECIMAL}[NSEW]*, *{RE_DECIMAL}[NSEW]*)",
            re.IGNORECASE
        )
    ),
    # line 7
    ("record.station_name",): ((words, ),
        ( 7,  (slice(40), ) )
    #("record.station_name", "_"): ((str, str),
            #re.compile("(.*) *(CGS)", re.IGNORECASE) 
        # )
    ),
    # line 8
    # ("record.channel", "record.component", "_", "record.station_channel", "record.location_name"): (
    ("record.channel", "record.component", "_", "_", "record.location_name"): (
        (str, str, maybe_t("(DegR*)",str), maybe_t("Sta Chn: ([0-9]*)", words), words),
        re.compile(# (  1   )   (---------)  (---)   (--)             (------)
            rf"Chan *([0-9]*): *([A-z0-9]*) *(DegR*)? *(.*) *Location: *([ -~]*)\s",
            re.IGNORECASE
        )
    ),
    # line 11
    ("record.instr_period", ".units"): ((float, units),
        re.compile(
            rf"Instr Period *= *({RE_DECIMAL}) *({RE_UNITS}),",
            re.IGNORECASE
        )
    ),
    #    # line 15
    #    (
    #        "filter.bandpass.point",
    #        "filter.bandpass.point.units",
    #        "filter.bandpass.limit_low",
    #        "filter.bandpass.limit_high",
    #        "filter.bandpass.limit.units"
    #    ) : (
    #        (float, units, float, float, units),
    #        re.compile(
    #            rf"Accelerogram bandpass filtered with *({RE_DECIMAL}) *({RE_UNITS}) pts at *({RE_DECIMAL}) and *({RE_DECIMAL}) *({RE_UNITS})\s",
    #            re.IGNORECASE
    #        )
    #    ),
    ("accel.peak_value", "accel.units", "accel.peak_time"): ((float, units, float),
        re.compile(
            rf"Peak *acceleration *= *({RE_DECIMAL}) *({RE_UNITS}) *at *({RE_DECIMAL})",
            re.IGNORECASE
        )
    ),
    ("veloc.peak_value", "veloc.units", "veloc.peak_time"): ((float, units, float),
        re.compile(
            rf"Peak *velocity *= *({RE_DECIMAL}) *({RE_UNITS}) *at *({RE_DECIMAL})",
            re.IGNORECASE
        )
    ),
    ("displ.peak_value", "displ.units", "displ.peak_time"): ((float, units, float),
        re.compile(
            rf"Peak *displacement *= *({RE_DECIMAL}) *({RE_UNITS}) *at *({RE_DECIMAL})",
            re.IGNORECASE
        )
    ),
    ("record.init_veloc", ".units", "record.init_displ", ".units"): ((float, units, float, units),
        re.compile(
            rf"Initial velocity *= *({RE_DECIMAL}) *({RE_UNITS}); *Initial displacement *= *({RE_DECIMAL}) *({RE_UNITS})\s",
            re.IGNORECASE
        )
    ),
    ("accel.shape", "accel.time_step"): ((int, float),
        re.compile(f"([0-9]*) *points of accel data equally spaced at *({RE_DECIMAL})", re.IGNORECASE)
    ),
    ("veloc.shape", "veloc.time_step"): ((int, float),
        re.compile(f"([0-9]*) *points of veloc data equally spaced at *({RE_DECIMAL})", re.IGNORECASE)
    ),
    ("displ.shape", "displ.time_step"): ((int, float),
        re.compile(f"([0-9]*) *points of displ data equally spaced at *({RE_DECIMAL})", re.IGNORECASE)
    ),
}
# fmt: on

V1_HEADER_FIELDS = HEADER_FIELDS.copy()
V1_HEADER_FIELDS.update({
    ("record.station_name",): ((words, ),
        ( 6,  (slice(40),) )
    )
})

def read_event(read_file, verbosity=0, summarize=False, **kwds):
    """
    Take the name of a CSMIP zip file and extract record data for the event.

    - kwds are passed to read_motion_v2
    """

    components = []
    zippath    = Path(read_file)
    archive    = zipfile.ZipFile(zippath)
    motions    = defaultdict(QuakeMotion)
    
    # Loop over V1 and V2 files in the zipped archive
    for file in archive.namelist():
        # Disregard any files that are not V1 or V2
        if not file.endswith((".v2", ".V2", ".v1", ".V1")):
            continue

        # Optional info logging
        if verbosity > 2: print(f"\t\t{file}", file=sys.stderr)

        v1 = True if file.endswith((".v1", ".V1")) else False

        cmp = read_record_v2(file, archive, verbosity=verbosity, summarize=summarize, v1=v1, **kwds)
        loc = _make_key(cmp.get("location_name", str(file)))
        drn = _make_key(cmp.get("component", "NA"))

        # Check that a Component with this direction has not already been
        # added to the Motion container under this location. This should
        # not happen, but is checked for robustness.
        if drn in motions[loc].components:
            loc += "_alt"

        motions[loc]["key"] = loc
        motions[loc].components[drn] = cmp


    # EVENT-LEVEL METADATA
    # Compute peak values over the entire archive.
    # --------------------------------------------
    # V1 files may not give peak values for the individual files (components), 
    # so in this case they are computed manually.
    if v1 and not summarize:
        peak_accel = max(
            (max(c.accel.data, key=abs) for m in motions.values() for c in m.components.values()),
            key=abs
        )

    # Otherwise, just take the peaks from the parsed metadata.
    else:
        peak_accel = max(
            (c.accel.get("peak_value", 0.0) for m in motions.values() for c in m.components.values()), 
            key=abs
        )

    # Collect someother information from the first file (component)
    first_motion    = list(motions.values())[0]
    first_component = list(first_motion.components.values())[0]

    date = first_component.get("date", "NA")
    metadata = {
        "file_name": str(read_file),
        "peak_accel": peak_accel,
        "event_date": date,
        "station_name":      first_component.get("station_name", "NA"),
        "station_coord":     first_component.get("station.coord", "NA"),
        "record_identifier": first_component.get("record_identifier", "NA"),
        "station_number":    first_component.get("station.no", "NA")
    }
    return QuakeCollection(dict(motions), event_date=date, meta=metadata)


# Fields that are not provided in the V1 format.
V1_EXCLUDE = ("filter*", "*peak*", "*init*", "*disp*", "*velo*")


def read_record_v2(
    read_file, 
    archive: zipfile.ZipFile = None,
    verbosity: int  = 0,
    summarize: bool =False,
    v1: bool = False,
    exclusions: tuple = (),
    **kwds
) -> QuakeComponent:
    """
    Read a ground motion record using the CSMIP Volume 2 format
    """
    if v1:
        exclusions = V1_EXCLUDE
        FIELDS = V1_HEADER_FIELDS
    else:
        FIELDS = HEADER_FIELDS

    filename = Path(read_file)

    # 1. PARSE READABLE HEADER (Regular expressions)
    # Collect keys to exclude
    keys = []
    for x in exclusions:
        for k in FIELDS:
            if any(fnmatch.fnmatch(kk,x) for kk in k):
                keys.append(k)

    header_fields = {k: v for k,v in FIELDS.items() if k not in keys}

    # Parse header fields
    try:
        with open_quake(read_file, "r", archive) as f:
            header_data = parse_sequential_fields(f, header_fields, verbose=verbosity)
        header_data.pop("_")
    except:
        if verbosity:
            print("Failed to parse header data for file {filename.name}", file=sys.stderr)
        header_data = {}


    # 2. PARSE NUMERIC HEADERS
    # Reopen and parse out data; Note, only the first call
    # provides a skip_header argument as successive reads
    # pick up where the previous left off.
    with open_quake(read_file, "r", archive) as f:
        # 50 integer values spanning 7 lines between lines 26-32
        int_header = np.genfromtxt(
            f,
            dtype=int,
            skip_header=13 if v1 else 25,
            max_rows=6,
            delimiter=5,
            ).flatten()
        int_header = np.append(int_header , np.genfromtxt(
            f,
            dtype=int,
            max_rows=1,
            delimiter=5,
            ).flatten()
        )
        assert len(int_header) == 100, int_header[:5]

        # 100 floating point values on lines 33-45
        real_header = np.genfromtxt(
            f,
            max_rows= 27-21 if v1 else 45-33,
            delimiter=10,
        ).flatten()
        real_header = np.append(real_header , np.genfromtxt(
            f,
            max_rows=1,
            delimiter=10,
            ).flatten()
        )

        assert len(real_header) == (50 if v1 else 100)
        
        # Clean and process numeric header data, setup for parse stage 3.
        num_header = _process_numeric_headers_v2(int_header, real_header, header_data)
        # extract information about shape of data
        s = next(f)
        s = s if isinstance(s, str) else s.decode("utf-8")
        len_accel = int(re.match("^ *([0-9]*) *.*", s).group(1))
        # extract format specifier, eg "(8f9.6)" if provided
        data_fmt = re.match(r"\(8f(.*)\)", s)
        if data_fmt:
            field_width = int(data_fmt.group(1).split(".")[0])
        else:
            field_width = 9 if v1 else 10

        parse_options = dict(
            delimiter = field_width,  # fields are 10 chars wide
            dtype=float,
        )

        # 3. PARSE OUT SENSOR DATA
        # Note that successive file reads will begin where we left off
        if not summarize:
            accel = np.genfromtxt(
                f,
                # skip_header=HEADER_END_LINE + 1,
                max_rows=np.ceil(len_accel/NUM_COLUMNS) - 1,
                **parse_options,
            ).flatten()
            accel = np.append(accel, np.genfromtxt(f,max_rows=1,**parse_options))
            if not v1:
                s = next(f)
                s = s if isinstance(s, str) else s.decode("utf-8")
                len_veloc = int(re.match("^ *([0-9]*)", s).group(0))
                veloc = np.genfromtxt(
                    f, max_rows=np.ceil(len_veloc/NUM_COLUMNS)-1, **parse_options
                ).flatten()
                veloc = np.append(veloc, np.genfromtxt(f,max_rows=1,**parse_options))

                s = next(f)
                s = s if isinstance(s, str) else s.decode("utf-8")
                len_displ = int(re.match("^ *([0-9]*)", s).group(0))
                displ = np.genfromtxt(
                    f, max_rows=np.ceil(len_displ/NUM_COLUMNS)-1, **parse_options
                ).flatten()
                displ = np.append(displ, np.genfromtxt(f,max_rows=1,**parse_options))
            else:
                veloc, displ = [], []
        else:
            accel, veloc, displ = [], [], []

    # Treat metadata
    try:
        filter_data = {
            key[16:]: header_data.pop(key)
            for key in [
              "filter.bandpass.point",
              "filter.bandpass.point.units",
              "filter.bandpass.limit_low",
              "filter.bandpass.limit_high",
              "filter.bandpass.limit.units"
            ]
        }
    except:
        filter_data = {}

    record_data = {"filters": [{"filter_type": "bandpass",**filter_data}]}
    series_data = defaultdict(dict)

    for key, val in header_data.items():
        typ, k = key.split(".", 1)
        if typ == "record":
            record_data.update({k: val})
        elif typ in ["accel", "veloc", "displ"]:
            series_data[typ].update({k: val})

    record_data["file_name"] = filename.name
    record_data["station_channel"] = str(int(re_digits.search(filename.name.split(".")[0]).group(0)))

    try:
        record_data.update({
            "peak_displ": series_data["displ"]["peak_value"],
            "peak_veloc": series_data["veloc"]["peak_value"],
            "peak_accel": series_data["accel"]["peak_value"]
        })
    except:
        pass

    # The raw numeric header data hasnt been too useful
    # record_data["ihdr"] = list(int_header)
    # record_data["rhdr"] = list(real_header)
    return QuakeComponent(
        QuakeSeries(accel, meta=series_data["accel"]),
        QuakeSeries(veloc, meta=series_data["veloc"]),
        QuakeSeries(displ, meta=series_data["displ"]),
        meta=record_data,
    )

def _process_numeric_headers_v2(ihdr, rhdr, txthdr):
    data = {}
    ref_azimuth = ihdr[32 -1]
    rel_orientation = ihdr[27 -1]
    #assert txthdr["veloc.shape"] == ihdr[64 - 1]
    #txthdr["record.earthquake_name"] = txthdr["record.earthquake_name"][:ihdr[29 - 1]+1]

    #assert ihdr[51 -1] == ihdr[52-1]
    #assert ihdr[28 -1] == ihdr[51-1]


# Declare file types handled by this module
FILE_TYPES = {
    "csmip.v1":  {"type": QuakeComponent,  "read": read_record_v2},
    "csmip.v2":  {"type": QuakeComponent,  "read": read_record_v2},
    "csmip.v3":  {"type": QuakeComponent,  "read": read_record_v2, "spec": ""},
    "csmip.zip": {"type": QuakeCollection, "read": read_event},
}


def read_record(read_file, *args):
    file = Path(read_file)
    if file.suffix.lower() == ".v2":
        return read_record_v2(file)

def read(read_file, input_format=None):
    pass




import re

import numpy as np

from quakeio.core import QuakeSeries
from quakeio.utils.parseutils import open_quake

RE_TIME_STEP = re.compile(r"DT=\s+(.*)SEC")


def read_nga(read_file, *args, **kwds):
    with open_quake(read_file, "r") as f:
        match = RE_TIME_STEP.search(next(f))
        header = [""]
        while RE_TIME_STEP.search(header[-1]) is None:
            header.append(next(f))
            
        #header = [next(f) for _ in range(4)]
    dt = float(RE_TIME_STEP.search(header[-1]).group(1).strip())
    #dt = float(match.group(1).strip())
    series_type = header[-2][:5].lower()
    accel = np.genfromtxt(read_file, skip_header=4, skip_footer=1).flatten()
    return QuakeSeries(
        accel, dict(time_step=dt, series_type=series_type, units="g")
    )


FILE_TYPES = {
    "nga.at2": {
        "read": read_nga,
        "type": QuakeSeries,
    }
}



import tempfile

import quakeio

from .test_csmip import test_read

def dict_equal(d1,d2):
    pass

def test_round_trip():
    motion = test_read()
    with tempfile.TemporaryFile("w+") as fp:
        quakeio.write(fp, motion, "json", summarize=True)
        fp.seek(0)
        motion_2 = quakeio.read(fp, "json")
    




# Claudio Perez
# Command line interface for QuakeIO tool
import argparse

import quakeio


def build_parser():
    # fmt: off
    parser = argparse.ArgumentParser(
        prog="quakeio", 
        description="""Parsers and utilities for processing and converting accelerograms.

        """,
        usage="quakeio [MODE] [OPTIONS] FILE"
    )
    modes = parser.add_argument_group("Modes")
    modes.add_argument("-A", "--all", dest="mode_process", action="store_true",
        help="Process all input data [default]"
    )
    modes.add_argument("-S", "--summarize", action="store_true",
        help="Omit time series data and superfluous fields from output"
    )
    modes.add_argument("-v", "--verbose", dest="verbosity", action="count", default=0,
        help="Increase output verbosity"
    )

    parser.add_argument("read_file", metavar="FILE", nargs="?", default="-")
    #parser.add_argument("-c", dest="command", action="append")
    #parser.add_argument("validate", action="store_true", default=False
    #)
    parser.add_argument(
        "-f",
        dest="input_format",
        metavar="FORMAT",
        help="Specify input file format",
        choices=list(quakeio.FILE_TYPES.keys()),
    )
    parser.add_argument(
        "-t", "--to",
        dest="write_format",
        metavar="FORMAT",
        help="Specify output file format",
        default="json",
    )
    parser.add_argument(
            "--human",
            default=False,
            action="store_true",
            help="Use human friendly field names."
    )
    parser.add_argument(
        "-x",
        "--exclude",
        dest="exclusions",
        metavar="KEY",
        help="Specify keys to exclude by parser",
        action="append",
        default=[]
    )
    parser.add_argument(
            "-m",
            dest="match_patterns",
            action="append",
            default=[]
    )
    parser.add_argument(
            "-o",
            dest="write_file",
            default="-",
            help="Specify an output file."
    )
    parser.add_argument(
        "-r",
        "--rotate",
        dest="angle",
        metavar="ANGLE",
        help="Rotate grouped records by ANGLE"
    )
    # parser.add_argument("-s", "--scale", help="Scale ground motions by factor FACTOR")

    parser.add_argument(
        "--version", help="Print version and exit.", action="version", version=quakeio.__version__
    )
    formats = parser.add_argument_group("Formats",
        description=f"""
        yaml, json, csmip, opensees
        """
        # {', '.join(quakeio.FILE_TYPES.keys())}
    )
    # fmt: on
    return parser


def cli(*args, write_file="-", human=False, validate=False, version=False, command=None, **kwds):
    read_file = kwds.pop("read_file")
    if "?" in read_file:
        from urllib.parse import urlparse
        url = urlparse(read_file)
        read_file = url.path
        query = url.query
    else:
        query = None

    motion = quakeio.read(read_file, **kwds)

    if kwds["match_patterns"]:
        kvs = {}
        for pat in kwds["match_patterns"]:
            key, val = pat.split("=")
            try:
                key, typ = key.split(":")
            except:
                typ = "l"
            kvs.update({key: val})
            motion = motion.match(typ, **kvs)
    if validate:
        run_validation(motion)
    if human:
        motion = quakeio.core.write_pretty(motion)
    if "angle" in kwds and kwds["angle"]:
        quakeio.core.rotate(motion, float(kwds["angle"]))

    quakeio.write(write_file, motion, **kwds)
    print("\n")

def run_validation(collection):
    for motion in collection.motions:
        for component in motion.components:
            for series in component.series:
                pass
    pass

def list_args(*args):
    args = build_parser().parse_args()
    return cli(**vars(args))

def main(args=None):
    args = build_parser().parse_args(args)
    try:
        cli(**vars(args))
    except Exception as e:
        print(args)
        raise e


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import setuptools

#import atexit
#from setuptools.command.install import install
#from setuptools.command.develop import develop
#from setuptools.command.egg_info import egg_info
#
#
#def install_styles():
#    try:
#        import matplotlib
#    except ImportError:
#        print("WARNING - Failed to install style sheets")
#        return
#    import shutil, os, glob
#
#    mpl_stylelib_dir = os.path.join(matplotlib.get_configdir(), "stylelib")
#    # Find all style files
#    stylefiles = glob.glob('styles/**/*.mplstyle', recursive=True)
#    # Find stylelib directory (where the *.mplstyle files go)
#    if not os.path.exists(mpl_stylelib_dir):
#        os.makedirs(mpl_stylelib_dir)
#    # Copy files over
#    print("Installing styles into", mpl_stylelib_dir)
#    for stylefile in stylefiles:
#        print(os.path.basename(stylefile))
#        shutil.copy(
#            stylefile,
#            os.path.join(mpl_stylelib_dir, os.path.basename(stylefile)))
#
#class InstallMoveFile(install):
#    def __init__(self, *args, **kwargs):
#        super().__init__(*args, **kwargs)
#        atexit.register(install_styles)
#
#class EggInfoMoveFile(egg_info):
#    def __init__(self, *args, **kwargs):
#        super().__init__(*args, **kwargs)
#        atexit.register(install_styles)
#
#class DevelopMoveFile(develop):
#    def __init__(self, *args, **kwargs):
#        super().__init__(*args, **kwargs)
#        atexit.register(install_styles)


if __name__ == "__main__":
    setuptools.setup(
#        cmdclass={
#            'install': InstallMoveFile,
#            'egg_info': EggInfoMoveFile,
#            'develop': DevelopMoveFile
#        }
    )




"""
1. find appropriate svg elements.
2. Add hover attribute to svg elements and event to call `show_motion_details`
"""
import xml.etree.ElementTree as ET

def read_host(filename):
    tree = ET.parse(filename)
    root = tree.getroot()
    pass

"""

{{ svg_string }}

<script>
{{ collection_json }}
</script>

<script>
function show_motion_details(channel_id) {

}
</script>
"""



import json
from copy import copy
from pathlib import Path

from quakeio.core import QuakeComponent, QuakeCollection, QuakeSeries
from quakeio.utils.parseutils import open_quake


class Unused:
    pass

def write_text(write_file, ground_motion, **kwds):
    import numpy
    if isinstance(ground_motion, QuakeComponent):
        dt = ground_motion.accel["time_step"]
        accel = numpy.asarray(ground_motion.accel.data)[:, None]
    elif isinstance(ground_motion, QuakeSeries):
        dt = ground_motion["time_step"]
        accel = numpy.asarray(ground_motion.data)[:, None]
    else:
        raise ValueError(
            f"Cannot conver motion of type `{type(ground_motion)}` to QuakeComponent"
        )
    with open_quake(write_file, "w") as f:
        numpy.savetxt(f, accel, fmt="%.8f", delimiter="\n")


class QuakeEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, "tolist"):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)

def read_basic(data:dict):
    if "motions" in data:
        for motion in data["motions"]:
            for component in motion["components"]:
                pass
    else:
        raise ValueError()
    pass

def create_series(data):
    series = data.pop("data")
    return QuakeSeries(series, meta=data)

def read_json(read_file, **kwds: Unused):

    return json.load(read_file)


def write_json(write_file, ground_motion, indent=4, summarize=False, **kwds):
    ground_motion = ground_motion.serialize(serialize_data=not summarize)
    try:
        with open_quake(write_file, "w") as f:
            json.dump(ground_motion, f, indent=indent, cls=QuakeEncoder)
    except BrokenPipeError:
        pass

def dumps_json(ground_motion, indent=4, summarize=False, **kwds):
    ground_motion = ground_motion.serialize(serialize_data=not summarize)
    return json.dumps(ground_motion, indent=indent, cls=QuakeEncoder)


def read_yaml(read_file, **k):
    import yaml

    return yaml.load(read_file, Loader=yaml.Loader)


def write_yaml(write_file, ground_motion, summarize=False, **kwds):
    import yaml

    ground_motion = ground_motion.serialize(serialize_data=not summarize)
    # if not summarize:
    #     ground_motion = ground_motion.serialize()

    with open_quake(write_file, "w") as f:
        yaml.dump(ground_motion, f)


FILE_TYPES = {
    "txt":  {"write": write_text},
    "json": {"read": read_json, "write": write_json},
    "yaml": {"read": read_yaml, "write": write_yaml},
    "quake.json": {"read": read_json, "write": write_json},
    "quake.yaml": {"read": read_yaml, "write": write_yaml},
}
